
""" Class description goes here. """

import logging
from yaml import load, dump, Loader, Dumper

from dataclay.commonruntime.Initializer import size_tracking
from dataclay.serialization.DataClaySerializable import DataClaySerializable
from dataclay.serialization.python.lang.StringWrapper import StringWrapper
import six

logger = logging.getLogger(__name__)


class ManagementMetaClass(type):

    def __init__(cls, name, bases, kwds):
        """This overlaps quite a bit with YAMLObjectMetaclass."""
        if name != "ManagementObject":
            yaml_tag = u"tag:yaml.org,2002:es.bsc.%s" % (cls.__module__)
            cls.yaml_loader = Loader
            cls.yaml_tag = yaml_tag  # used by `ManagementObject.to_yaml`
            logger.trace("YAML TAG : %s", yaml_tag)
            Loader.add_constructor(yaml_tag, cls.from_yaml)
            Dumper.add_representer(cls, cls.to_yaml)
        super(ManagementMetaClass, cls).__init__(name, bases, kwds)

    def __new__(cls, name, bases, dct):
        if "_fields" not in dct:
            raise AttributeError("All YAML structures must have a `_fields` list attribute")

        if name == "ManagementObject":
            dct["__slots__"] = tuple()
            return super(ManagementMetaClass, cls).__new__(cls, name, bases, dct)

        # Note that ManagementObject class instantiation cannot reach here,
        # or a cyclic requirement will happen and it will crash

        if ManagementObject not in bases:
            # We are instantiating a derived class, merge all the attributes as expected
            full_fields = list()
            full_internal = list()

            for b in bases:
                if issubclass(b, ManagementObject):
                    # Note that we are not resolving correctly overrides
                    # But we do NOT expect the override on typed fields
                    # (it would be quite pathological)
                    try:
                        dct["_typed_fields"].update(b._typed_fields)
                    except KeyError:
                        pass

                    full_fields += b._fields
                    try:
                        full_internal += b._internal_fields
                    except AttributeError:
                        pass

            full_fields += dct["_fields"]
            if "_internal_fields" in dct:
                full_internal += dct["_internal_fields"]

            # Ensure that the generated class has complete `_fields` and `_internal_fields`
            dct["_fields"] = full_fields
            dct["_internal_fields"] = full_internal
            all_fields = full_fields + full_internal

        else:
            all_fields = list(dct["_fields"])

            if "_internal_fields" in dct:
                all_fields += dct["_internal_fields"]

        dct["__slots__"] = tuple(all_fields)

        return super(ManagementMetaClass, cls).__new__(cls, name, bases, dct)


@six.add_metaclass(ManagementMetaClass)
class ManagementObject(DataClaySerializable):
    _fields = list()
    _internal_fields = list()
    _typed_fields = dict()

    yaml_tag = None

    def __init__(self, **kwargs):
        super(ManagementObject, self).__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_yaml(cls, loader, node):
        """See YAMLObject."""
        return loader.construct_yaml_object(node, cls)

    @classmethod
    def to_yaml(cls, dumper, data):
        """See YAMLObject."""
        return dumper.represent_yaml_object(cls.yaml_tag, data, cls)

    def __getstate__(self):
        """This method is used both by Pickle and by YAML Dumper."""

        # All required fields should be there
        ret = {k: getattr(self, k, None) for k in self._fields}

        # Include all internal fields which are present (even if they are None)
        ret.update({k: getattr(self, k) for k in self._internal_fields
                    if hasattr(self, k)})
        return ret

    def __setstate__(self, state):
        """This method is used both by Pickle and by YAML Loader."""
        setted_attrs = set()
        unsetted_attrs = set()
        for k, v in state.items():
            # TODO: Check better fields

            if isinstance(v, dict) and k in self._typed_fields:
                typed_field = self._typed_fields[k]
                new_attr = typed_field()
                new_attr.__setstate__(v)
                v = new_attr

            try:
                setattr(self, k, v)
                setted_attrs.add(k)
            except Exception:
                unsetted_attrs.add(k)

        missed_fields = set(self._fields) - setted_attrs

        if len(missed_fields) != 0:
            logger.error("WARNING -- __setstate__ on class %s called without this fields: %s" 
            , self.__class__, list(missed_fields))
        if len(unsetted_attrs) > 0:
            logger.error("WARNING -- Attributes %s are not setted -- Fields missing on class %s"
            , unsetted_attrs, self.__class__)

    def __str__(self):
        lines = ["ManagementObject: %s" % self.__class__.__name__]
        for field_name in self._fields:
            try:
                lines.append("  - %s: %r" % (field_name, getattr(self, field_name)))
            except AttributeError:
                logger.debug("WARNING -- Missing attribute: %s", field_name)
                pass

        return "\n".join(lines)

    def serialize(self, io_file):
        """Serialize this instance into a IO like (file, StringIO...)."""
        # This is roughly equivalent to write into a string
        # and perform a Str().write, but with less memory copies
        with size_tracking(io_file):
            dump(self, io_file, encoding='utf-16-be', Dumper=Dumper)

    @classmethod
    def deserialize(cls, io_file):
        """Deserialize the IO into a new instance."""
        value = StringWrapper().read(io_file)
        return load(value, Loader=Loader)

    read = deserialize

    @classmethod
    def write(cls, io_file, value):
        assert isinstance(value, cls), \
            "Called `write` on class '%s' with an object of class '%s'" % (
                cls.__name__, type(value).__name__)
        value.serialize(io_file)

    @staticmethod
    def type_equal(deserialized, orig_val):
        ret_bool = True
        ret_list = []
        try:
            if deserialized.typeName == orig_val.typeName:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("typeName")
            pass
        try:
            if deserialized.languageDepInfos == orig_val.languageDepInfos:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("languageDepInfos")
            pass
        try:
            if deserialized.descriptor == orig_val.descriptor:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("descriptor")
            pass

        try:
            if deserialized.signature == orig_val.signature:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("signature")
            pass
        try:
            if deserialized.includes == orig_val.includes:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("includes")
            pass

        return ret_bool, ret_list

    @staticmethod
    def prop_equal(deserialized, orig_val, i=None):
        ret_bool = True
        ret_list = []
        # for i in range(0, len(orig_val)):
        if i is not None:
            try:
                if deserialized.namespace == orig_val[i].namespace:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append(str(i) + " namespace")

            try:
                if deserialized.className == orig_val[i].className:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append(str(i) + " className")

            try:
                if deserialized.getterImplementationID == orig_val[i].getterImplementationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append(str(i) + " getterImplementationID")

            try:
                if deserialized.setterImplementationID == orig_val[i].setterImplementationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " setterImplementationID")

            try:
                if deserialized.name == orig_val[i].name:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " name")

            try:
                if deserialized.position == orig_val[i].position:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " position")

            try:
                if deserialized.getterOperationID == orig_val[i].getterOperationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " getterOperationID")

            try:
                if deserialized.setterOperationID == orig_val[i].setterOperationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " setterOperationID")

            try:
                if deserialized.metaClassID == orig_val[i].metaClassID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " metaClassID")

            try:
                if deserialized.namespaceID == orig_val[i].namespaceID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append(str(i) + " namespaceID")

            try:
                if deserialized.languageDepInfos == orig_val[i].languageDepInfos:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append(str(i) + " languageDepInfos")
        else:
            # MetaClass_Factory
            try:
                if deserialized.namespace == orig_val.namespace:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append("namespace")

            try:
                if deserialized.className == orig_val.className:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append("className")

            try:
                if deserialized.getterImplementationID == orig_val.getterImplementationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append("getterImplementationID")

            try:
                if deserialized.setterImplementationID == orig_val.setterImplementationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("setterImplementationID")

            try:
                if deserialized.name == orig_val.name:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("name")

            try:
                if deserialized.position == orig_val.position:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("position")

            try:
                if deserialized.getterOperationID == orig_val.getterOperationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("getterOperationID")

            try:
                if deserialized.setterOperationID == orig_val.setterOperationID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("setterOperationID")

            try:
                if deserialized.metaClassID == orig_val.metaClassID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("metaClassID")

            try:
                if deserialized.namespaceID == orig_val.namespaceID:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("namespaceID")

            try:
                if deserialized.languageDepInfos == orig_val.languageDepInfos:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list

            except AttributeError:
                ret_list.append("languageDepInfos")

        return ret_bool, ret_list

    @staticmethod
    def op_equal(deserialized, orig_val, b_par=False):
        ret_bool = True
        ret_list = []
        try:
            if deserialized.namespace == orig_val.namespace:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("namespace")

        try:
            if deserialized.className == orig_val.className:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("className")

        try:
            if deserialized.descriptor == orig_val.descriptor:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("descriptor")

        try:
            if deserialized.signature == orig_val.signature:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("signature")

        try:
            if deserialized.name == orig_val.name:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("name")

        try:
            if deserialized.nameAndDescriptor == orig_val.nameAndDescriptor:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("nameAndDescriptor")
        if b_par:
            # Not executed by Metaclass_Factory
            try:
                if deserialized.params == orig_val.params:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("params")

            try:
                if deserialized.returnType == orig_val.returnType:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("returnType")

            try:
                if deserialized.implementations == orig_val.implementations:
                    ret_bool = True
                else:
                    ret_bool = False
                    return ret_bool, ret_list
            except AttributeError:
                ret_list.append("implementations")

        try:
            if deserialized.paramsOrder == orig_val.paramsOrder:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("paramsOrder")

        try:
            if deserialized.isAbstract == orig_val.isAbstract:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("isAbstract")

        try:
            if deserialized.isStaticConstructor == orig_val.isStaticConstructor:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("isStaticConstructor")

        try:
            if deserialized.metaClassID == orig_val.metaClassID:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("metaClassID")

        try:
            if deserialized.namespaceID == orig_val.namespaceID:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("namespaceID")

        try:
            if deserialized.languageDepInfos == orig_val.languageDepInfos:
                ret_bool = True
            else:
                ret_bool = False
                return ret_bool, ret_list
        except AttributeError:
            ret_list.append("languageDepInfos")

        return ret_bool, ret_list




""" Class description goes here. """

"""gRPC ExecutionEnvironment Client code - StorageLocation/EE methods."""

import logging
import sys
import traceback
import six
import grpc

if six.PY2:
    import cPickle as pickle
elif six.PY3:
    import _pickle as pickle

from grpc._cython.cygrpc import ChannelArgKey
from dataclay.commonruntime.Settings import settings
from dataclay.communication.grpc import Utils
from dataclay.communication.grpc.generated.dataservice import dataservice_pb2_grpc
from dataclay.communication.grpc.messages.dataservice import dataservice_messages_pb2
from dataclay.exceptions.exceptions import DataClayException
from dataclay.util.YamlParser import dataclay_yaml_dump
import dataclay.communication.grpc.messages.common.common_messages_pb2 as CommonMessages
from dataclay.util import Configuration

__author__ = 'Enrico La Sala <enrico.lasala@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


class EEClient(object):

    def __init__(self, hostname, port):
        """Create the stub and the channel at the address passed by the server."""
        self.address = str(hostname) + ":" + str(port)
        options = [(ChannelArgKey.max_send_message_length, -1),
                   (ChannelArgKey.max_receive_message_length, -1)]

        self.channel = grpc.insecure_channel(self.address, options)
        try:
            grpc.channel_ready_future(self.channel).result(timeout=Configuration.GRPC_CHECK_ALIVE_TIMEOUT)
        except Exception as e:
            sys.exit('Error connecting to server %s' % self.address)
        else:
            self.ds_stub = dataservice_pb2_grpc.DataServiceStub(self.channel)

    def close(self):
        """Closing channel by deleting channel and stub"""
        del self.channel
        del self.ds_stub
        self.channel = None
        self.ds_stub = None

    def ds_deploy_metaclasses(self, namespace_name, deployment_pack):
        deployment_pack_dict = dict()

        for k, v in deployment_pack.items():
            deployment_pack_dict[k] = dataclay_yaml_dump(v)

        request = dataservice_messages_pb2.DeployMetaClassesRequest(
            namespace=namespace_name,
            deploymentPack=deployment_pack_dict
        )

        try:
            response = self.ds_stub.deployMetaClasses(request)

        except RuntimeError as e:
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def ds_new_persistent_instance(self, session_id, class_id, implementation_id, i_face_bitmaps, params):

        logger.debug("Ready to call to a DS to build a new persistent instance for class {%s}",
                     class_id)
        temp_iface_b = dict()
        temp_param = None

        if i_face_bitmaps is not None:
            for k, v in i_face_bitmaps.items():
                temp_iface_b[k] = v

        if params is not None:
            temp_param = Utils.get_param_or_return(params)

        request = dataservice_messages_pb2.NewPersistentInstanceRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            classID=Utils.get_msg_options['meta_class'](class_id),
            implementationID=Utils.get_msg_options['implem'](implementation_id),
            ifaceBitMaps=temp_iface_b,
            params=temp_param
        )

        try:
            response = self.ds_stub.newPersistentInstance(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        return Utils.get_id(response.objectID)
    
    def ds_store_objects(self, session_id, objects, moving, ids_with_alias):

        obj_list = []
        id_with_alias_list = []

        for obj in objects:
            obj_list.append(Utils.get_obj_with_data_param_or_return(obj))

        if ids_with_alias is not None:
            for id_with_alias in ids_with_alias:
                if id_with_alias is not None:
                    id_with_alias_list.append(Utils.get_msg_options['object'](id_with_alias))

        request = dataservice_messages_pb2.StoreObjectsRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objects=obj_list,
            moving=moving,
            idsWithAlias=id_with_alias_list
        )

        try:
            response = self.ds_stub.storeObjects(request)

        except RuntimeError as e:
            traceback.print_exc(file=sys.stdout)
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def ds_new_metadata(self, md_infos):

        md_infos_dict = dict()

        for k, v in md_infos.items():

            md_infos_dict[k] = dataclay_yaml_dump(v)

        request = dataservice_messages_pb2.NewMetaDataRequest(
            mdInfos=md_infos_dict
        )

        try:
            response = self.ds_stub.newMetaData(request)

        except RuntimeError as e:
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def ds_get_copy_of_object(self, session_id, object_id, recursive):
        request = dataservice_messages_pb2.GetCopyOfObjectRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectID=Utils.get_msg_options['object'](object_id),
            recursive=recursive,
        )
        
        try:
            response = self.ds_stub.getCopyOfObject(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)
        
        serialized_obj = Utils.get_param_or_return(response.ret)
        
        return serialized_obj
        
    def ds_update_object(self, session_id, into_object_id, from_object):
        request = dataservice_messages_pb2.UpdateObjectRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            intoObjectID=Utils.get_msg_options['object'](into_object_id),
            fromObject=Utils.get_param_or_return(from_object)
        )
        
        try:
            response = self.ds_stub.updateObject(request)

        except RuntimeError as e:
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def ds_get_objects(self, session_id, object_ids, recursive, moving):

        object_ids_list = []
        for oid in object_ids:
            object_ids_list.append(Utils.get_msg_options['object'](oid))

        request = dataservice_messages_pb2.GetObjectsRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectIDS=object_ids_list,
            recursive=recursive,
            moving=moving
        )

        try:
            response = self.ds_stub.getObjects(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)
        
        serialized_objs = list() 
        for obj_with_data in response.objects:
            serialized_objs.append(Utils.get_obj_with_data_param_or_return(obj_with_data))
        
        return serialized_objs
    
    def get_referenced_objects_ids(self, session_id, object_ids):
        
        object_ids_list = []
        for oid in object_ids:
            object_ids_list.append(Utils.get_msg_options['object'](oid))
        
        request = dataservice_messages_pb2.GetReferencedObjectIDsRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectIDS=object_ids_list
        )

        try:
            response = self.ds_stub.getReferencedObjectsIDs(request)
        
        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)
        
        objs_ids = list() 
        for oid in response.objectIDs:
            objs_ids.append(Utils.get_id(oid))
        
        return objs_ids

    def ds_get_federated_objects(self, dc_instance_id, object_ids):

        object_ids_list = []
        for oid in object_ids:
            object_ids_list.append(Utils.get_msg_options['object'](oid))

        request = dataservice_messages_pb2.GetFederatedObjectsRequest(
            extDataClayID=Utils.get_msg_options['dataclay_instance'](dc_instance_id),
            objectIDS=object_ids_list
        )

        try:
            response = self.ds_stub.getFederatedObjects(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)
        
        serialized_objs = list() 
        for obj_with_data in response.objectsList:
            serialized_objs.append(Utils.get_obj_with_data_param_or_return(obj_with_data))
        
        return serialized_objs

    def ds_new_version(self, session_id, object_id, metadata_info):
        
        logger.info("This new_version is called somewhere?")

        request = dataservice_messages_pb2.NewVersionRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectID=Utils.get_msg_options['object'](object_id),
            metadataInfo=dataclay_yaml_dump(metadata_info)
        )

        try:
            response = self.ds_stub.newVersion(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        result = dict()
        oid = Utils.get_id(response.objectID)

        for k, v in response.versionedIDs:
            result[Utils.get_id_from_uuid(k)] = Utils.get_id_from_uuid(v)

        t = (oid, result)

        return t

    def ds_consolidate_version(self, session_id, version_info):

        request = dataservice_messages_pb2.ConsolidateVersionRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            versionInfo=dataclay_yaml_dump(version_info)
        )

        try:
            response = self.ds_stub.consolidateVersion(request)

        except RuntimeError as e:
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def ds_upsert_objects(self, session_id, object_bytes):

        obj_byt_list = []
        for entry in object_bytes:
            obj_byt_list.append(Utils.get_obj_with_data_param_or_return(entry))

        request = dataservice_messages_pb2.UpsertObjectsRequest(
                sessionID=Utils.get_msg_options['session'](session_id),
                bytesUpdate=obj_byt_list)

        try:
            response = self.ds_stub.upsertObjects(request)

        except RuntimeError as e:
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def make_persistent(self, session_id, params):
        logger.debug("Client performing MakePersistent")

        request = dataservice_messages_pb2.MakePersistentRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            params=Utils.get_param_or_return(params),
        )

        try:
            response = self.ds_stub.makePersistent(request)

        except RuntimeError as e:
            logger.error('Failed to make persistent', exc_info=True)
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)
        
    def federate(self, session_id, params):

        request = dataservice_messages_pb2.FederateRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            params=Utils.get_param_or_return(params),
        )

        try:
            response = self.ds_stub.federate(request)

        except RuntimeError as e:
            logger.error('Failed to federate', exc_info=True)
            raise e

        if response.isException:
            raise DataClayException(response.exceptionMessage)
        
    def ds_execute_implementation(self, object_id, implementation_id, session_id, params):
        logger.debug("Client performing ExecuteImplementation")

        request = dataservice_messages_pb2.ExecuteImplementationRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            implementationID=Utils.get_msg_options['implem'](implementation_id),
            params=Utils.get_param_or_return(params),
            objectID=Utils.get_msg_options['object'](object_id)
        )

        try:
            response = self.ds_stub.executeImplementation(request)

        except RuntimeError as e:
            logger.error('Failed to execute implementation', exc_info=True)
            raise e

        if response.excInfo.isException:
            try:
                exception = pickle.loads(response.excInfo.serializedException)
            except:
                raise DataClayException(response.excInfo.exceptionMessage)
            else:
                raise exception

        if response.ret is not None:
            return Utils.get_param_or_return(response.ret)
        else:
            return None
        
    def ds_new_replica(self, session_id, object_id, recursive):

        request = dataservice_messages_pb2.NewReplicaRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectID=Utils.get_msg_options['object'](object_id),
            recursive=recursive
        )

        try:
            response = self.ds_stub.newReplica(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        result = set()

        for oid in response.replicatedIDs:
            result.add(Utils.get_id(oid))

        return result

    def ds_move_objects(self, session_id, object_id, dest_st_location, recursive):

        request = dataservice_messages_pb2.MoveObjectsRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectID=Utils.get_msg_options['object'](object_id),
            destLocID=Utils.get_msg_options['storage_loc'](dest_st_location),
            recursive=recursive
        )

        try:
            response = self.ds_stub.moveObjects(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        result = set()

        for oid in response.movedObjects:
            result.add(Utils.get_id(oid))

        return result

    def ds_remove_objects(self, session_id, object_ids, recursive, moving, new_hint):

        obj_ids_list = []
        for oid in object_ids:
            obj_ids_list.append(Utils.get_msg_options['object'](oid))

        request = dataservice_messages_pb2.RemoveObjectsRequest(
            sessionID=Utils.get_msg_options['session'](session_id),
            objectIDs=obj_ids_list,
            recursive=recursive,
            moving=moving,
            newHint=Utils.get_msg_options['exec_env'](new_hint)
        )

        try:
            response = self.ds_stub.removeObjects(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        result = dict()

        for k, v in response.removedObjects.items():
            result[Utils.get_id_from_uuid(k)] = Utils.get_id_from_uuid(v)

        return result

    def ds_migrate_objects_to_backends(self, back_ends):

        back_ends_dict = dict()

        for k, v in back_ends.items():
            back_ends_dict[k] = dataclay_yaml_dump(v)

        request = dataservice_messages_pb2.MigrateObjectsRequest(
            destStorageLocs=back_ends_dict
        )

        try:
            response = self.ds_stub.migrateObjectsToBackends(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        result = dict()

        for k, v in response.migratedObjs.items():
            m_objs = v
            oids = set()

            for oid in m_objs.getObjsList():
                oids.add(Utils.get_id(oid))

            result[Utils.get_id_from_uuid(k)] = oids

        non_migrated = set()

        for oid in response.nonMigratedObjs.getObjsList():
            non_migrated.add(Utils.get_id(oid))

        t = (result, non_migrated)

        return t
    
    # STORAGE LOCATION - DBHANDLER

    def update_refs(self, ref_counting):
        
        """ ref_counting is a dict uuid - integer """ 
        request = dataservice_messages_pb2.UpdateRefsRequest(
            refsToUpdate=ref_counting
        )

        try:
            response = self.ds_stub.updateRefs(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def store_to_db(self, execution_environment_id, object_id, obj_bytes):
        
        request = dataservice_messages_pb2.StoreToDBRequest(
            executionEnvironmentID=Utils.get_msg_id_backend_id(execution_environment_id),
            objectID=Utils.get_msg_options['object'](object_id),
            objBytes=obj_bytes
        )

        try:
            response = self.ds_stub.storeToDB(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def get_from_db(self, execution_environment_id, object_id):

        request = dataservice_messages_pb2.GetFromDBRequest(
            executionEnvironmentID=Utils.get_msg_id_backend_id(execution_environment_id),
            objectID=Utils.get_msg_options['object'](object_id),
        )
        try:
            response = self.ds_stub.getFromDB(request)

        except RuntimeError as e:
            raise e

        if response.excInfo.isException:
            raise DataClayException(response.excInfo.exceptionMessage)

        return response.objBytes
    
    def update_to_db(self, execution_environment_id, object_id, new_obj_bytes, dirty):
        request = dataservice_messages_pb2.UpdateToDBRequest(
            executionEnvironmentID=Utils.get_msg_id_backend_id(execution_environment_id),
            objectID=Utils.get_msg_options['object'](object_id),
            objBytes=new_obj_bytes,
            dirty=dirty
        )

        try:
            response = self.ds_stub.updateToDB(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)
    
    def delete_to_db(self, execution_environment_id, object_id):
        request = dataservice_messages_pb2.DeleteToDBRequest(
            executionEnvironmentID=Utils.get_msg_id_backend_id(execution_environment_id),
            objectID=Utils.get_msg_options['object'](object_id),
        )

        try:
            response = self.ds_stub.deleteToDB(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)
        
    def associate_execution_environment(self, execution_environment_id):
        request = dataservice_messages_pb2.AssociateExecutionEnvironmentRequest(
            executionEnvironmentID=Utils.get_msg_id_backend_id(execution_environment_id)
        )

        try:
            response = self.ds_stub.associateExecutionEnvironment(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def activate_tracing(self, task_id):
        request = dataservice_messages_pb2.ActivateTracingRequest(
            taskid=task_id
        )
        
        try:
            response = self.ds_stub.activateTracing(request)
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)

    def deactivate_tracing(self):
        try:
            response = self.ds_stub.deactivateTracing(CommonMessages.EmptyMessage())
        
        except RuntimeError as e:
            raise e
        
        if response.isException:
            raise DataClayException(response.exceptionMessage)
        
    def get_traces(self):
        try:
            response = self.ds_stub.getTraces(CommonMessages.EmptyMessage())
        except RuntimeError as e:
            raise e
        
        result = dict()
        for k, v in response.stubs.items():
            result[k] = v

        return result
        
    ################## EXTRAE IGNORED FUNCTIONS ###########################
    deactivate_tracing.do_not_trace = True
    activate_tracing.do_not_trace = True




""" Class description goes here. """

"""gRPC ExecutionEnvironment Server code - StorageLocation/EE methods."""

from io import BytesIO
import logging
import traceback
import six

if six.PY2:
    import cPickle as pickle
elif six.PY3:
    import _pickle as pickle

from dataclay.commonruntime.Runtime import getRuntime
from dataclay.communication.grpc import Utils
from dataclay.communication.grpc.generated.dataservice import dataservice_pb2_grpc as ds
from dataclay.communication.grpc.messages.common import common_messages_pb2
from dataclay.communication.grpc.messages.dataservice import dataservice_messages_pb2
from dataclay.exceptions.exceptions import DataClayException
from dataclay.util.YamlParser import dataclay_yaml_load

__author__ = 'Enrico La Sala <enrico.lasala@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)

_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class DataServiceEE(ds.DataServiceServicer):

    interceptor = None

    def __init__(self, theexec_env, interceptor=None):
        """ Execution environment being managed """
        self.execution_environment = theexec_env
        DataServiceEE.interceptor = interceptor
        
    def ass_client(self):
        self.client = getRuntime().ready_clients["@STORAGE"]

    def get_exception_info(self, ex):
        ex_message = None
        if hasattr(ex, "message"):
            ex_message = ex.message
            logger.warning("Exception produced with message:\n%s", ex_message)

        try:
            ex_serialized = pickle.dumps(ex)
        except TypeError:
            logger.warning("Could not serialize %s", ex)
            ex_serialized = None

        return common_messages_pb2.ExceptionInfo(
                isException=True,
                serializedException=ex_serialized,
                exceptionMessage=Utils.prepare_exception(ex_message, Utils.return_stack())
            )

    def deployMetaClasses(self, request, context):
        
        logger.verbose("[deployMetaClasses] Deploying classes")

        try:
            namespace = request.namespace
            classes_map_yamls = request.deploymentPack 
            self.execution_environment.ds_deploy_metaclasses(namespace, classes_map_yamls)
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def newPersistentInstance(self, request, context):

        try:
            iface_bit_maps = {}

            for k, v in request.ifaceBitMaps.items():
                iface_bit_maps[Utils.get_id_from_uuid(k)] = Utils.prepare_bytes(v)

            params = []

            if request.params:
                params = Utils.get_param_or_return(request.params)

            oid = self.client.ds_new_persistent_instance(Utils.get_id(request.sessionID),
                                                         Utils.get_id(request.classID),
                                                         Utils.get_id(request.implementationID),
                                                         iface_bit_maps,
                                                         params)

            return dataservice_messages_pb2.NewPersistentInstanceResponse(objectID=Utils.get_msg_options['object'](oid))

        except Exception as ex:
            return dataservice_messages_pb2.NewPersistentInstanceResponse(
                excInfo=self.get_exception_info(ex))

    def storeObjects(self, request, context):
                
        try:
            objects_list = []
            for vol_param in request.objects:
                param = Utils.get_obj_with_data_param_or_return(vol_param)
                serialized_obj = BytesIO(param[3])
                dcobj = param[0], param[1], param[2], serialized_obj
                objects_list.append(dcobj)

            ids_with_alias_set = set()
            session_id = Utils.get_id(request.sessionID)
            if request.idsWithAlias is not None and len(request.idsWithAlias) > 0:
                for ids_with_alias in request.idsWithAlias:
                    ids_with_alias_set.add(Utils.get_id(ids_with_alias))

            self.execution_environment.store_objects(session_id, objects_list, request.moving, ids_with_alias_set)
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)
    
    def makePersistent(self, request, context):
        try:
            objects_to_persist = Utils.get_param_or_return(request.params)
            session_id = Utils.get_id(request.sessionID)
            self.execution_environment.make_persistent(session_id, objects_to_persist)
            logger.verbose("MakePersistent finished, sending response")
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            traceback.print_exc()
            return self.get_exception_info(ex)
            
    def federate(self, request, context):
        try:
            logger.verbose("Federation started")
            objects_to_persist = Utils.get_param_or_return(request.params)
            session_id = Utils.get_id(request.sessionID)
            self.execution_environment.federate(session_id, objects_to_persist)
            logger.verbose("Federation finished, sending response")
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def unfederate(self, request, context):
        try:
            logger.verbose("Unfederation started")
            session_id = Utils.get_id(request.sessionID)
            object_ids = set()
            for oid in request.objectIDs:
                object_ids.add(Utils.get_id(oid))
            self.execution_environment.unfederate(session_id, object_ids)
            logger.verbose("Unfederation finished, sending response")
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            traceback.print_exc()
            return self.get_exception_info(ex)

    def executeImplementation(self, request, context):
        logger.debug("Starting ExecuteImplementation handling")

        try:
            object_id = Utils.get_id(request.objectID)
            implementation_id = Utils.get_id(request.implementationID)
            serialized_params = Utils.get_param_or_return(request.params)
            session_id = Utils.get_id(request.sessionID)
            result = self.execution_environment.ds_exec_impl(object_id, implementation_id, serialized_params, session_id)
            logger.verbose("ExecuteImplementation finished, sending response")

            return dataservice_messages_pb2.ExecuteImplementationResponse(ret=Utils.get_param_or_return(result))

        except Exception as ex:
            traceback.print_exc()
            return dataservice_messages_pb2.ExecuteImplementationResponse(
                excInfo=self.get_exception_info(ex))

    def getCopyOfObject(self, request, context):
        try:
            result = self.execution_environment.get_copy_of_object(
                Utils.get_id(request.sessionID),
                Utils.get_id(request.objectID),
                request.recursive)
            
            return dataservice_messages_pb2.GetCopyOfObjectResponse(ret=Utils.get_param_or_return(result))
        
        except Exception as ex:
            return dataservice_messages_pb2.GetObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )
            
    def updateObject(self, request, context):
        try:
            
            self.execution_environment.update_object(
                Utils.get_id(request.sessionID),
                Utils.get_id(request.intoObjectID),
                Utils.get_param_or_return(request.fromObject)
                )
            
            logger.verbose("updateObject finished, sending response")
            
            return common_messages_pb2.ExceptionInfo()

        except DataClayException as ex:
            return self.get_exception_info(ex)

    def getObjects(self, request, context):
        try:
            object_ids = set()
            for oid in request.objectIDS:
                object_ids.add(Utils.get_id(oid))

            result = self.execution_environment.get_objects(Utils.get_id(request.sessionID),
                                            object_ids,
                                            request.recursive,
                                            request.moving)

            obj_list = []

            for entry in result:
                obj_list.append(Utils.get_obj_with_data_param_or_return(entry))

            return dataservice_messages_pb2.GetObjectsResponse(objects=obj_list)

        except Exception as ex:
            return dataservice_messages_pb2.GetObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )

    def getReferencedObjectsIDs(self, request, context):
        try:
            object_ids = set()
            for oid in request.objectIDS:
                object_ids.add(Utils.get_id(oid))
            
            result = self.execution_environment.get_referenced_objects_ids(
                                            Utils.get_id(request.sessionID),
                                            object_ids)
            obj_list = []
            for entry in result:
                obj_list.append(Utils.get_msg_options['object'](entry))
    
            return dataservice_messages_pb2.GetReferencedObjectIDsResponse(objectIDs=obj_list)

        except Exception as ex:
            return dataservice_messages_pb2.GetReferencedObjectIDsResponse(
                excInfo=self.get_exception_info(ex)
            )

    def getFederatedObjects(self, request, context):
        try:
            object_ids = set()
            for oid in request.objectIDS:
                object_ids.add(Utils.get_id(oid))
            
            result = self.execution_environment.get_federated_objects(
                Utils.get_id(request.extDataClayID),
                object_ids
            )
            obj_list = []

            for entry in result:
                obj_list.append(Utils.get_obj_with_data_param_or_return(entry))
    
            return dataservice_messages_pb2.GetFederatedObjectsResponse(objects=obj_list)

        except DataClayException as ex:
            return dataservice_messages_pb2.GetFederatedObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )

    def newMetaData(self, request, context):

        try:
            md_infos = {}

            for k, v in request.mdInfos.items():

                md_infos[Utils.get_id_from_uuid(k)] = dataclay_yaml_load(v)

            self.client.ds_new_metadata(md_infos)

            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def newVersion(self, request, context):

        try:
            result = self.execution_environment.new_version(
                Utils.get_id(request.sessionID),
                Utils.get_id(request.objectID),
                dataclay_yaml_load(request.metadataInfo)
            )

            vers_ids = dict()

            for k, v in result[1].items():
                vers_ids[Utils.prepare_bytes(str(k))] = Utils.prepare_bytes(str(v))

            return dataservice_messages_pb2.NewVersionResponse(
                objectID=Utils.get_msg_options['object'](result[0]),
                versionedIDs=vers_ids
            )

        except Exception as ex:
            return dataservice_messages_pb2.NewVersionResponse(
                excInfo=self.get_exception_info(ex)
            )

    def consolidateVersion(self, request, context):

        try:
            self.execution_environment.consolidate_version(Utils.get_id(request.sessionID),
                                           dataclay_yaml_load(request.versionInfo))

            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def upsertObjects(self, request, context):

        try:
            session_id = Utils.get_id(request.sessionID)

            objects = []
            for entry in request.bytesUpdate:
                objects.append(Utils.get_obj_with_data_param_or_return(entry))

            self.execution_environment.upsert_objects(session_id, objects)

            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def newReplica(self, request, context):

        ##### NOT WELL IMPL. NEW ALGORITHM MISSING FOR MOVING OBJECTS CHECKING FIRST THE MEMORY ######

        try:
            result = self.execution_environment.new_replica(Utils.get_id(request.sessionID),
                                            Utils.get_id(request.objectID),
                                            request.recursive)

            repl_ids_list = []

            for oid in result:
                repl_ids_list.append(Utils.get_msg_options['object'](oid))

            return dataservice_messages_pb2.NewReplicaResponse(
                replicatedIDs=repl_ids_list
            )

        except Exception as ex:
            traceback.print_exc()
            return dataservice_messages_pb2.NewReplicaResponse(
                excInfo=self.get_exception_info(ex)
            )

    def moveObjects(self, request, context):

        ##### NOT WELL IMPL. NEW ALGORITHM MISSING FOR MOVING OBJECTS CHECKING FIRST THE MEMORY ######

        try:
            result = self.execution_environment.move_objects(Utils.get_id(request.sessionID),
                                                 Utils.get_id(request.objectID),
                                                 Utils.get_id(request.destLocID),
                                                 request.recursive)

            mov_obj_list = []

            for oid in result:
                mov_obj_list.append(Utils.get_msg_options['object'](oid))

            return dataservice_messages_pb2.MoveObjectsResponse(
                movedObjects=mov_obj_list
            )

        except Exception as ex:
            return dataservice_messages_pb2.MoveObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )

    def removeObjects(self, request, context):

        try:
            object_ids = set()

            for oid in request.getObjectsIDSList():
                object_ids.add(Utils.get_id(oid))

            result = self.client.ds_remove_objects(Utils.get_id(request.sessionID),
                                                   object_ids,
                                                   request.recursive,
                                                   request.moving,
                                                   Utils.get_id(request.newHint))

            rem_obj = dict()

            for k, v in result.items():
                rem_obj[str(k)] = str(v)

            return dataservice_messages_pb2.RemoveObjectsResponse(
                removedObjects=rem_obj
            )

        except Exception as ex:
            return dataservice_messages_pb2.RemoveObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )
    
    def exists(self, request, context):
        try: 
            exists = self.execution_environment.exists(Utils.get_id(request.objectID))
            return dataservice_messages_pb2.ExistsResponse(
                exists=exists
            )
        except Exception as ex:
            return dataservice_messages_pb2.ExistsResponse(
                excInfo=self.get_exception_info(ex)
            )
    
    def updateRefs(self, request, context):
        try: 

            """ deserialize into dictionary of object id - integer """ 
            ref_counting = dict() 
            for serialized_oid, counter in request.refsToUpdate.items():
                ref_counting[serialized_oid] = counter

            self.execution_environment.update_refs(ref_counting)
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            traceback.print_exc()
            return self.get_exception_info(ex)
      
    def getRetainedReferences(self, request, context):
        try: 

            result = self.execution_environment.get_retained_references()
            retained_refs = []

            for oid in result:
                retained_refs.append(Utils.get_msg_options['object'](oid))
            return dataservice_messages_pb2.GetRetainedReferencesResponse(retainedReferences=retained_refs)

        except Exception as ex:
            return self.get_exception_info(ex)

    def closeSessionInDS(self, request, context):
        try: 
            self.execution_environment.close_session_in_ee(Utils.get_id(request.sessionID))
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def migrateObjectsToBackends(self, request, context):

        try:
            backends = dict()

            for k, v in request.destStorageLocs.items():
                backends[Utils.get_id_from_uuid(k)] = dataclay_yaml_load(v)

            result = self.client.ds_migrate_objects_to_backends(backends)

            migr_obj_res = dict()

            for k, v in result[0].items():
                migrated_obj_list = list()

                for oid in v:
                    migrated_obj_list.append(Utils.get_msg_options['object'](oid))
                
                migrated_obj_builder = dataservice_messages_pb2.MigratedObjects(objs=migrated_obj_list)
                migr_obj_res[str(k)] = migrated_obj_builder

            non_migrated_objs_list = list()

            for oid in result[1]:
                non_migrated_objs_list.append(Utils.get_msg_options['object'](oid))
            
            non_migrated_objs_builder = dataservice_messages_pb2.MigratedObjects(objs=non_migrated_objs_list)

            return dataservice_messages_pb2.MigrateObjectsResponse(
                migratedObjs=migr_obj_res,
                nonMigratedObjs=non_migrated_objs_builder
            )

        except Exception as ex:
            return dataservice_messages_pb2.MigrateObjectsResponse(
                excInfo=self.get_exception_info(ex)
            )

    def activateTracing(self, request, context):
        try: 

            self.execution_environment.activate_tracing(request.taskid)
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)

    def getTraces(self, request, context):
        try: 
            result = self.execution_environment.get_traces()
            return common_messages_pb2.GetTracesResponse(traces=result)
        except Exception as ex:
            return self.get_exception_info(ex)
    
    def deactivateTracing(self, request, context):
        try:
            self.execution_environment.deactivate_tracing()
            return common_messages_pb2.ExceptionInfo()

        except Exception as ex:
            return self.get_exception_info(ex)




""" Class description goes here. """

'''
Created on 26 ene. 2018

@author: dgasull
'''
from dataclay.heap.HeapManager import HeapManager
from threading import Timer


class ClientHeapManager(HeapManager):
    """
    @summary: This class is intended to manage all dataClay objects in Client runtime's memory.
    """
    
    def __init__(self, theruntime):
        """
        @postcondition: Constructor of the object 
        @param theruntime: Runtime being managed 
        """ 
        HeapManager.__init__(self, theruntime)
        self.logger.debug("CLIENT HEAP MANAGER created")
            
    def add_to_heap(self, dc_object):
        """
        @postcondition: the object is added to dataClay's heap
        @param dc_object: object to add to the heap 
        """
        HeapManager._add_to_inmemory_map(self, dc_object)     
        
    def flush_all(self): 
        """ 
        @postcondition: Does nothing. This function is intended to be used only in Execution Environment. Since 
        Heap Manager is an abstract class, we define this method as an "empty" method. Maybe I am missing some 
        better way to do that? (dgasull)
        """
        return
    
    def run_task(self): 
        self.cleanReferencesAndLockers()




""" Class description goes here. """

"""Utilities to deploy/save Source files into the filesystem.

Ensuring that the folder hierarchy and the __init__.py files are in their place
is done in this module.
"""


import logging
import os.path
from dataclay.commonruntime.Initializer import logger

__author__ = ['Alex Barcelo <alex.barcelo@bsc.es']
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

# Almost hardcoded, but easily reachable... quite hacky, I know
MAGIC_LINE_NUMBER_FOR_IMPORTS = 14


def _ensure_package(package_path, ds_deploy):
    """Create the folder package_path and create the __init__.py file inside."""
    if not os.path.exists(package_path):
        os.mkdir(package_path)

    init_file = os.path.join(package_path, "__init__.py")
    if not os.path.exists(init_file):
        # logger.debug("Creating __init__.py file at folder %s", package_path)
        with open(init_file, 'w') as f:
            f.writelines([
                u"# Automatically created by dataClay\n",
                u"import os\n",
                u"import sys\n",
                u"\n",
                u"from dataclay import dclayMethod, dclayEmptyMethod, DataClayObject\n",
                u"from dataclay.contrib.dummy_pycompss import *\n",
                u"\n",
                u"from logging import getLogger\n",
                u"getLogger('dataclay.deployed_classes').debug('Package path: %s')\n" % package_path,
                u"StorageObject = DataClayObject\n",
                u"\n",
                u"###############################################################\n",
                u"######### <Class-based imports>:\n",
                u"\n",
                # Here the lines will be inserted, eventually, as needed (14th line at the moment)
                u"\n",
                u"######### </Class-based imports>\n",
                u"###############################################################\n",
            ])
            # ATTENTION! If you add any line to the previous header, update the variable
            # MAGIC_LINE_NUMBER_FOR_IMPORTS

            if not ds_deploy:
                # Client may require the PyCOMPSs imports
                # the fallback is already set up (dummy_pycompss module)
                f.writelines([
                    u"\n",
                    u"try:\n",
                    u"    from pycompss.api.task import task\n",
                    u"    from pycompss.api.parameter import *\n",
                    u"    from pycompss.api.constraint import constraint\n",
                    u"except ImportError:\n",
                    u"    pass\n",
                    u"\n",
                ])


def deploy_class(namespace, full_name, source, imports, source_deploy_path, ds_deploy=False):
    """Deploy a class source to the filesystem.

    :param namespace: The namespace of the class (first part of package name).
    :param full_name: The full name (including package) of the class.
    :param source: The Python's source code for the class.
    :param source_deploy_path: The root for deployed source code.
    :param ds_deploy: If true, that means that the deploy is for a DataService (not client)
    """
    _ensure_package(source_deploy_path, ds_deploy)

    package, klass = ("%s.%s" % (namespace, full_name)).rsplit('.', 1)
    logger.info("Going to deploy class %s in path %s/__init__.py",
                klass, package.replace(".", "/"))

    current_path = source_deploy_path
    for p in package.split("."):
        current_path = os.path.join(current_path, p)
        _ensure_package(current_path, ds_deploy)

    class_path = os.path.join(current_path, "__init__.py")
    logger.info("Class destination file: %s", class_path)

    if not os.path.exists(class_path):
        raise IOError("__init__.py file in package %s should have been already initialized"
                      % package)
    else:
        # Because we need to insert imports in a nice place, read & write approach
        with open(class_path, 'rt') as f:
            contents = f.readlines()

        contents = contents[:MAGIC_LINE_NUMBER_FOR_IMPORTS] \
                   + [imports] \
                   + contents[MAGIC_LINE_NUMBER_FOR_IMPORTS:] \
                   + [source]
        
        with open(class_path, 'wt') as f:
            f.write("".join(contents))




""" Class description goes here. """

from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class BooleanWrapper(DataClayPythonWrapper):
    """One-byte bool type (0 means False)."""
    __slots__ = ()

    def __init__(self):
        pass

    def read(self, io_file):
        val = IntegerWrapper(8).read(io_file)
        if val == 0:
            return False
        else:
            return True

    def write(self, io_file, value):
        if value:
            IntegerWrapper(8).write(io_file, 0x01)
        else:
            IntegerWrapper(8).write(io_file, 0x00)





""" Class description goes here. """

from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
import logging
import os

from dataclay.commonruntime.Runtime import getRuntime
from dataclay.commonruntime.Settings import settings
from dataclay.util.FileUtils import deploy_class
from dataclay.util.YamlParser import Loader, dataclay_yaml_load

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

_babel_data = None
logger = logging.getLogger(__name__)


def babel_stubs_load(stream):
    # Note the explicit list, just in case the caller wants to close the provided file/buffer
    map_babel_stubs = dataclay_yaml_load(stream, Loader=Loader)
    result = list()
    for k, v in map_babel_stubs.items():
        result.append(dataclay_yaml_load(v))
    return result


def prepare_storage(stubs_folder=None):
    """Ensure (force creation if not exists) the STUB_STORAGE folder.

    By default, the settings.stubs_folder is used. You can override by providing
    the stubs_folder argument.
    """
    if not stubs_folder:
        stubs_folder = settings.stubs_folder

    if not os.path.exists(stubs_folder):
        os.mkdir(stubs_folder)

    if not os.path.isdir(stubs_folder):
        raise IOError("The `StubsClasspath` is not a folder --check file and permissions")


def load_babel_data(stubs_folder=None):
    """Load all Babel Stub data from the cached file.

    By default, the settings.stubs_folder is used. You can override by providing
    the stubs_folder argument.

    :return: A dictionary (the parsed YAML).
    """
    global _babel_data
    if _babel_data is None:
        with open(os.path.join(stubs_folder or settings.stubs_folder,
                               "babelstubs.yml"),
                  'rb') as f:
            _babel_data = babel_stubs_load(f)

    return _babel_data


def deploy_stubs(stubs_folder=None):
    """Perform the actual deployment of classes (python files).

    By default, the settings.stubs_folder is used. You can override by providing
    the stubs_folder argument.
    """
    if not stubs_folder:
        stubs_folder = settings.stubs_folder

    # Use the stored StubInfo (which is a YAML)
    babel_data = load_babel_data(stubs_folder)
    source_deploy = os.path.join(stubs_folder, 'sources')
    try:
        os.makedirs(source_deploy)
    except OSError as e:
        if e.errno != 17:
            # Not the "File exists" expected error, reraise it
            raise

    for class_data in babel_data:
        namespace = class_data.namespace
        full_name = class_data.className

        logger.debug("Deploying stub for %s::%s", namespace, full_name)
    
        try:
            # ToDo: to avoid clashes, in the future, maybe the full name should include the namespace
            # like: ... stubs_folder, "%s.%s" % (namespace, full_name ...
            with open(os.path.join(stubs_folder, full_name), 'rt') as f:
                source = f.read()
        except IOError:
            source = ""
    
        deploy_class(namespace, full_name, source, u"", source_deploy)


def track_local_available_classes():
    """Track the available classes into the commonruntime.local_available_classes.

    Note that no deployment is done in this function: the deployment should be
    done beforehand through the deploy_stubs function.

    This function returns all the contracts that have been found.
    """
    babel_data = load_babel_data()
    contracts = set()

    for class_data in babel_data:
        contracts.update(class_data.contracts)
        namespace = class_data.namespace
        full_name = class_data.className

        getRuntime().local_available_classes[class_data.classID] = \
            "%s.%s" % (namespace, full_name)

    logger.verbose("Using the following contracts: %s", contracts)
    return contracts




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class DataSet(ManagementObject):
    _fields = ["dataClayID",
               "name"
               ]

    _internal_fields = ["providerAccountID",
                        ]



"""Entry point for easy dump to get extrae values per method and class module. 

This module can be run as a standalone and will dump a file to be read from all python nodes 
(must be same event number in all nodes) to be added into PCF paraver file. 
"""

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

import ast
import os

if __name__ == "__main__":
    source_dir = "%s/src" % str(os.getcwd())
    current_dir = "%s/dataclay" % source_dir
    dest_dir = "%s/python_paraver_values.properties" % (os.getcwd())
    final_file = open(dest_dir, 'w') 
    current_method_value = 10000
    for root, dirname, filename in os.walk(current_dir):
        pth_build = ""
        if os.path.isfile(root + "/__init__.py"):
            for i in filename:
                if i != "__init__.py" and i != "__init__.pyc":
                    if i.split('.')[1] == "py":
                        slot = list(set(root.split('\\')) - set(os.getcwd().split('\\')))
                        pth_build = slot[0]
                        del slot[0]
                        for j in slot:
                            pth_build = pth_build + "/" + j
                        
                        class_name = i.split('.')[0]
                        module_dir = pth_build + "/" + i
                        module_name = module_dir.replace("/", ".").replace(".py", "")
                        # print pth_build + "." + i.split('.')[0]
                        with open(module_dir) as file:
                            node = ast.parse(file.read())
                        
                        classes = [n for n in node.body if isinstance(n, ast.ClassDef)]
                        for class_ in classes:
                            methods = [n for n in class_.body if isinstance(n, ast.FunctionDef)]
                            for method in methods:
                                actual_module_name = module_dir.replace(source_dir, "").replace("/", ".").replace(".py", "")[1:]
                                actual_method = actual_module_name + "." + class_.name + "." + method.name
                                final_file.write(actual_method + "=" + str(current_method_value) + "\n")
                                current_method_value = current_method_value + 1

    final_file.close()
    



""" Class description goes here. """
import importlib
import uuid
import six
import os
import logging
import traceback
from abc import ABCMeta, abstractmethod
from logging import TRACE
from lru import LRU
from grpc import RpcError
from dataclay.heap.LockerPool import LockerPool
from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.paraver import extrae_tracing_is_enabled, initialize_extrae, get_current_available_task_id, \
    finish_tracing, get_task_id
from dataclay.communication.grpc.clients.ExecutionEnvGrpcClient import EEClient
from dataclay.communication.grpc.clients.LogicModuleGrpcClient import LMClient
from dataclay.commonruntime.Runtime import threadLocal
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton
from dataclay.serialization.lib.SerializationLibUtils import SerializationLibUtilsSingleton
from dataclay.exceptions.exceptions import DataClayException
from dataclay.util import Configuration

""" Make this class abstract """


@six.add_metaclass(ABCMeta)
class DataClayRuntime(object):
        
    """ Logger """ 
    logger = logging.getLogger('dataclay.api')

    def __init__(self):
        """ Cache of alias """
        # TODO: un-hardcode this
        self.alias_cache = LRU(50000)
        
        """ Cache of EE info """
        self.ee_info_map = None

        """ GRPC clients """
        self.ready_clients = dict()
        
        """ Cache of classes. TODO: is it used? -> Yes, in StubUtils and ClientObjectLoader"""
        self.local_available_classes = dict()
        
        """  Heap manager. Since it is abstract it must be initialized by sub-classes. 
        DataClay-Java uses abstract functions to get the field in the proper type (EE or client) 
        due to type-check. Not needed here. """
        self.dataclay_heap_manager = None
        
        """ Object loader. """
        self.dataclay_object_loader = None
        
        """  Locker Pool in runtime. This pool is used to provide thread-safe implementations in dataClay. """
        self.locker_pool = LockerPool()
        
        """ Indicates if runtime was initialized. TODO: check if same in dataclay.api -> NO """
        self.__initialized = False
        
        """ Indicates volatiles being send - to avoid race-conditions """
        self.volatile_parameters_being_send = set()
        
        # Local info of thread
        self.thread_local_info = threadLocal
        
        """ Cache of metadata """
        self.metadata_cache = LRU(50)
        
    @abstractmethod
    def initialize_runtime_aux(self): pass
    
    def initialize_runtime(self):
        """ 
        IMPORTANT: getRuntime can be called from decorators, during imports, and therefore a runtime might be created. 
        In that case we do NOT want to create threads to start. Only if "init" was called (client side) or 
        server was started. This function is for that.
        """
        self.logger.debug("INITIALIZING RUNTIME")
        self.initialize_runtime_aux()
        self.dataclay_heap_manager.start()

    def is_initialized(self):
        """
        @return: TRUE if runtime is initialized (Client 'init', EE should be always TRUE). False otherwise.
        """
        return self.__initialized
        
    @abstractmethod
    def is_exec_env(self):
        """
        @return: TRUE if runtime is for EE. False otherwise.
        """
        pass
    
    @abstractmethod
    def get_session_id(self):
        """ Get session id.
        :returns: session id
        :rtype: uuid
        """
        pass
    
    def get_lm_api(self, host, port):
        """ Get logic module connection.
        :param host: logic module host
        :param port: logic module port 
        :type host: str
        :type port: str
        :returns: logic module api to connect
        :rtype: logic module grpc api
        """
        try:
            lm_client = self.ready_clients[host + ":" + str(port)]
        except KeyError:
            lm_client = LMClient(host, port)
            try: 
                lm_client.check_alive()
            except: 
                raise 
            self.ready_clients[host + ":" + str(port)] = lm_client
        return lm_client 
    
    @abstractmethod
    def get_operation_info(self, object_id, operation_name): pass
  
    @abstractmethod
    def get_implementation_id(self, object_id, operation_name, implementation_idx=0): pass
    
    def get_object_by_id(self, object_id, class_id=None, hint=None):
        """ Get object instance directly from an object id, use class id and hint in 
        case it is still not registered.
        :param object_id: id of the object to get 
        :param class_id: class id of the object to get 
        :param hint: hint of the object to get
        :returns: object instance
        :rtype: DataClayObject
        :type object_id: uuid
        :type class_id: uuid
        :type hint: uuid
        :rtype: DataClayObject
        """
        self.logger.debug("Get object %s by id", str(object_id))
        o = self.get_from_heap(object_id)
        if o is not None:
            return o

        if not class_id:
            full_name, namespace = self.ready_clients["@LM"].get_object_info(
                self.get_session_id(), object_id)
            self.logger.debug("Trying to import full_name: %s from namespace %s",
                         full_name, namespace)

            # Rearrange the division, full_name may include dots (and be nested)
            prefix, class_name = ("%s.%s" % (namespace, full_name)).rsplit('.', 1)
            m = importlib.import_module(prefix)
            klass = getattr(m, class_name)
            class_id = klass.get_class_extradata().class_id

        o = self.get_or_new_persistent_instance(object_id, class_id, hint)
        return o

    def add_to_heap(self, dc_object):
        """
        @postcondition: the object is added to dataClay's heap
        @param dc_object: object to add to the heap 
        """
        self.dataclay_heap_manager.add_to_heap(dc_object)
    
    def update_object_id(self, old_object_id, new_object_id):
        """
        @postcondition: the object references are updated into dataClay's heap
        @param old_object_id: the old id of the object
        @param new_object_id: the new id of the object
        """
        self.dataclay_heap_manager.update_object_id(old_object_id, new_object_id)
    
    def remove_from_heap(self, object_id):
        """
        @postcondition: Remove reference from Heap. Even if we remove it from the heap, 
        the object won't be Garbage collected till HeapManager flushes the object and releases it.
        @param object_id: id of object to remove from heap
        """
        self.dataclay_heap_manager.remove_from_heap(object_id)
        
    def get_or_new_volatile_instance_and_load(self, object_id, metaclass_id, hint,
                                              obj_with_data, ifacebitmaps):
        """
        @postcondition: Get from Heap or create a new volatile in EE and load data on it.
        @param object_id: ID of object to get or create 
        @param metaclass_id: ID of class of the object (needed for creating it) 
        @param hint: Hint of the object, can be None. 
        @param obj_with_data: data of the volatile instance 
        @param ifacebitmaps: interface bitmaps
        """ 
        return self.dataclay_object_loader.get_or_new_volatile_instance_and_load(metaclass_id, object_id, hint, obj_with_data, ifacebitmaps)
    
    def add_volatiles_under_deserialization(self, volatiles):
        """
        @postcondition: Add volatiles provided to be 'under deserialization' in case any execution in a volatile is thrown 
        before it is completely deserialized. This is needed in case any deserialization depends on another (not for race conditions)
        like hashcodes or other similar cases.
        @param volatiles: volatiles under deserialization
        """
        self.thread_local_info.volatiles_under_deserialitzation = volatiles
   
    def remove_volatiles_under_deserialization(self):
        """
        @postcondition: Remove volatiles under deserialization
        """
        self.thread_local_info.volatiles_under_deserialitzation = None
        
    def get_copy_of_object(self, from_object, recursive):
        session_id = self.get_session_id()
        
        backend_id = from_object.get_location()
        try:
            execution_client = self.ready_clients[backend_id]
        except KeyError:
            exeenv = self.get_execution_environments_info()[backend_id]
            execution_client = EEClient(exeenv.hostname, exeenv.port)
            self.ready_clients[backend_id] = execution_client
        
        copiedObject = execution_client.ds_get_copy_of_object(session_id, from_object.get_object_id(), recursive)
        result = DeserializationLibUtilsSingleton.deserialize_params_or_return(copiedObject, None, None, None, self)
            
        return result[0]            
        
    def update_object(self, into_object, from_object):
        session_id = self.get_session_id()
        
        backend_id = into_object.get_location()
        try:
            execution_client = self.ready_clients[backend_id]
        except KeyError:
            exeenv = self.get_execution_environments_info()[backend_id]
            execution_client = EEClient(exeenv.hostname, exeenv.port)
            self.ready_clients[backend_id] = execution_client
        
        # We serialize objects like volatile parameters
        parameters = list()
        parameters.append(from_object)
        # TODO: modify serialize_params_or_return to not require this
        params_order = list()
        params_order.append("object")
        params_spec = dict()
        params_spec["object"] = "DataClayObject"  # not used, see serialized_params_or_return
        ser_from = SerializationLibUtilsSingleton.serialize_params_or_return(
            params=parameters,
            iface_bitmaps=None,
            params_spec=params_spec,
            params_order=params_order,
            hint_volatiles=backend_id,
            runtime=self,
            recursive=True,
            for_update=True)
        
        vol_objects = ser_from[3]
        if vol_objects is not None:
            new_ids = dict()
            
            for tag in vol_objects:
                cur_oid = ser_from[3][tag][0]
                if cur_oid not in new_ids:
                    if cur_oid == from_object.get_object_id():
                        new_ids[cur_oid] = into_object.get_object_id()
                    else:
                        new_ids[cur_oid] = uuid.uuid4()
                
                ser_from[3][tag] = (new_ids[cur_oid],) + ser_from[3][tag][1:]
            
            for vol_tag in vol_objects:
                oids = ser_from[3][vol_tag][2][0]
                for tag, oid in oids.items():
                    if oid in new_ids:
                        try:
                            ser_from[3][vol_tag][2][0][tag] = new_ids[oid]
                        except KeyError: 
                            pass
        
        self.logger.debug("Sending updated metadata: %s", str(ser_from))

        execution_client.ds_update_object(session_id, into_object.get_object_id(), ser_from)
        
    def get_from_heap(self, object_id):
        """
        @postcondition: Get from heap. 
        @param object_id: id of object to get from heap
        @return Object with id provided in heap or None if not found.
        """
        return self.dataclay_heap_manager.get_from_heap(object_id)
    
    def lock(self, object_id):
        """
        @postcondition: Lock object with ID provided
        @param object_id: ID of object to lock 
        """
        self.locker_pool.lock(object_id)
        
    def unlock(self, object_id):
        """
        @postcondition: Unlock object with ID provided
        @param object_id: ID of object to unlock 
        """
        self.locker_pool.unlock(object_id)    
        
    def run_remote(self, object_id, backend_id, operation_name, value):
        session_id = self.get_session_id()
        implementation_id = self.get_implementation_id(object_id, operation_name)

        try:
            execution_client = self.ready_clients[backend_id]
        except KeyError:
            exeenv = self.get_execution_environments_info()[backend_id]
            execution_client = EEClient(exeenv.hostname, exeenv.port)
            self.ready_clients[backend_id] = execution_client

        operation = self.get_operation_info(object_id, operation_name)
        value = SerializationLibUtilsSingleton.serialize_params_or_return(
            params=(value,),
            iface_bitmaps=None,
            params_spec=operation.params,
            params_order=operation.paramsOrder,
            hint_volatiles=None,
            runtime=self
        )

        ret = execution_client.ds_execute_implementation(object_id, implementation_id,
            session_id, value)

        if ret is not None:
            self.logger.trace("Execution return %s of type %s", ret, operation.returnType)
            return DeserializationLibUtilsSingleton.deserialize_return(ret, None, operation.returnType, self)
        
    def call_execute_to_ds(self, instance, parameters, operation_name, exeenv_id, using_hint):
        
        object_id = instance.get_object_id()
        operation = self.get_operation_info(object_id, operation_name)
        session_id = self.get_session_id()
        implementation_id = self.get_implementation_id(object_id, operation_name)

        # // === SERIALIZE PARAMETERS === //
        serialized_params = SerializationLibUtilsSingleton.serialize_params_or_return(
            params=parameters,
            iface_bitmaps=None,
            params_spec=operation.params,
            params_order=operation.paramsOrder,
            hint_volatiles=exeenv_id,
            runtime=self)
        
        if serialized_params is not None and serialized_params[3] is not None:
            for param in serialized_params[3].values():
                self.volatile_parameters_being_send.add(param[0])
            
        # // === EXECUTE === //
        max_retry = Configuration.MAX_EXECUTION_RETRIES
        num_misses = 0
        executed = False
        for _ in range(max_retry):
            try:
                self.logger.verbose("Obtaining API for remote execution in %s ", exeenv_id)
                execution_client = self.ready_clients[exeenv_id]
            except KeyError:
                exeenv = self.get_execution_environments_info()[exeenv_id] 
                self.logger.debug("Not found in cache ExecutionEnvironment {%s}! Starting it at %s:%d",
                               exeenv_id, exeenv.hostname, exeenv.port)
                execution_client = EEClient(exeenv.hostname, exeenv.port)
                self.ready_clients[exeenv_id] = execution_client
    
            try:
                self.logger.verbose("Calling remote EE %s ", exeenv_id)
                ret = execution_client.ds_execute_implementation(
                    object_id,
                    implementation_id,
                    session_id,
                    serialized_params)
                executed = True
                break
            
            except (DataClayException, RpcError) as dce:
                self.logger.warning("Execution resulted in an error, retrying...", exc_info=dce)

                is_race_condition = False
                if serialized_params is not None and serialized_params[4] is not None:
                    for param in serialized_params[4]:
                        if param.get_object_id() in self.volatile_parameters_being_send:
                            is_race_condition = True
                            break
                if not is_race_condition:
                    num_misses = num_misses + 1
                    self.logger.debug("Exception dataclay during execution. Retrying...")
                    self.logger.debug(str(dce))
                    if object_id in self.metadata_cache:   
                        del self.metadata_cache[object_id]

                    try:
                        metadata = self.get_metadata(object_id)
                        new_location = False
                    except DataClayException:
                        metadata = None
    
                    if metadata is None:
                        self.logger.warning("Execution failed and no metadata available. Cannot continue")
                        raise
    
                    for loc in metadata.locations:
                        self.logger.debug("Found location %s" % str(loc))
                        if loc != exeenv_id:
                            exeenv_id = loc
                            self.logger.debug("Found different location %s" % str(loc))
                            new_location = True
                            break
                        
                    if not new_location: 
                        exeenv_id = next(iter(metadata.locations))
                
                    if using_hint:
                        instance.set_hint(exeenv_id)
                    
                    self.logger.debug("[==Miss Jump==] MISS. The object %s was not in the exec.location %s. Retrying execution." 
                                % (instance.get_object_id(), str(exeenv_id)));
    
        self.logger.verbose("Result of operation named '%s' received", operation_name)
        if serialized_params is not None and serialized_params[3] is not None:
            for param in serialized_params[3].values():
                if num_misses > 0: 
                    #===========================================================
                    # if there was a miss, it means that the persistent object in which we were executing 
                    # was not in the choosen location. As you can see in the serialize parameters function above
                    # we provide the execution environment as hint to set to volatile parameters. In EE, before
                    # deserialization of volatiles we check if the persistent object in which to execute a method is
                    # there, if not, EE raises and exception. Therefore, if there was a miss, we know that the 
                    # hint we set in volatile parameters is wrong, because they are going to be deserialized/stored
                    # in the same location as the object with the method to execute
                    #===========================================================
                    self.logger.debug("Looking in Heap for parameter %s", str(param[0]))
                    param_instance = self.get_from_heap(param[0])
                    param_instance.set_hint(exeenv_id)
                self.volatile_parameters_being_send.remove(param[0])
    
        if not executed: 
            raise RuntimeError("[dataClay] ERROR: Trying to execute remotely object  but something went wrong. "
                               "Maybe the object is still not stored (in case of asynchronous makepersistent) and "
                               "waiting time is not enough. Maybe the object does not exist anymore due to a remove. "
                               "Or Maybe an exception happened in the server and the call failed.")
    
        if ret is None:
            return None
        else:
            return DeserializationLibUtilsSingleton.deserialize_return(ret, None, operation.returnType, self)
    
    def synchronize_federated(self, instance, params, operation_name, dc_info_id):
        self.logger.debug("Calling external dataClay to run %s operation in object %s"
                          , operation_name, instance.get_object_id())
        operation = self.get_operation_info(instance.get_object_id(), operation_name)
        implementation_id = self.get_implementation_id(instance.get_object_id(), operation_name)
        # === SERIALIZE PARAMETERS === 
        # Between DC - DC , ifaceBitMaps = null
        serialized_params = SerializationLibUtilsSingleton.serialize_params_or_return(
            params=params,
            iface_bitmaps=None,
            params_spec=operation.params,
            params_order=operation.paramsOrder,
            hint_volatiles=instance.get_hint(),
            runtime=self)
        
        dc_info = self.get_external_dataclay_info(dc_info_id)
        hosts = dc_info.hosts 
        ports = dc_info.ports 
        for i in range(0, len(hosts)):
            try:
                lm_client = self.get_lm_api(hosts[i], ports[i])
                self.logger.debug("[==JUMP==] Request execution to external dataClay %s with host %s and port %s for object %s"
                                  , dc_info_id, hosts[i], ports[i], instance.get_object_id())

                lm_client.synchronize_federated_object(self.get_dataclay_id(), instance.get_object_id(),
                                               implementation_id, serialized_params)     
            except:
                traceback.print_exc()
                if i + 1 == len(hosts):
                    raise RuntimeError("[dataClay] ERROR: Cannot connect to external dataClay with ID %s" % (str(dc_info_id)))
        
    def ensure_object_is_registered(self, session_id, object_id, class_id, hint):
        """ 
        Ensure registration of an object. new replica/version/consolidate/move
        algorithms should not require registered metadata in LogicModule since
        new make persistent implementation behaves like volatiles and metadata
        is created eventually, not synchronously. Currently, we try to register
        it and if it is already registered, just continue.
        :param session_id: ID of session registering object
        :param object_id: ID of object to register
        :param class_id: ID of class of the object
        :param hint: Hint of the object
        :returns: None
        :type session_id: dataClay ID
        :type object_id: dataClay ID
        :type class_id: dataClay ID
        :type hint: dataClay ID
        :rtype: None
        """
        # FIXME: new replica/version/consolidate/move algorithms should not require
        # registered metadata in LogicModule since new make persistent implementation
        # behaves like volatiles and metadata is created eventually, not synchronously.
        # Currently, we try to register it and if it is already registered, just continue.
        # Make sure object is registered.
        # DataSet is None since it is obtained from session at LM.
        reg_info = [object_id, class_id, session_id, None]
        # In next call, alias must be null
        # NOTE: LogicModule register object function does not return an exception for already registered
        # object. We should never call registerObject for already registered objects and that's dataClay
        # code (check isPendingToRegister in EE or isPersistent,.. see makePersistent), and remember that,
        # this is a workaround, registerObject should never be called for replica/version/consolidate algorithms,
        # we must change the algorithms to not depend on metadata.
        # Also, location in which to register the object is the hint (in case it is not registered yet).
        self.ready_clients["@LM"].register_object(reg_info, hint, None, LANG_PYTHON)
    
    def federate_object(self, object_id, ext_dataclay_id, recursive, class_id, hint):
        session_id = self.get_session_id()
        self.logger.debug("[==FederateObject==] Starting federation of object %s with dataClay %s, and session %s", object_id, ext_dataclay_id, session_id)
        self.ensure_object_is_registered(session_id, object_id, class_id, hint)
        self.ready_clients["@LM"].federate_object(session_id, object_id, ext_dataclay_id, recursive)
    
    def unfederate_object(self, object_id, ext_dataclay_id, recursive):
        session_id = self.get_session_id()
        self.logger.debug("[==UnfederateObject==] Starting unfederation of object %s with dataClay %s, and session %s", object_id, ext_dataclay_id, session_id)
        self.ready_clients["@LM"].unfederate_object(session_id, object_id, ext_dataclay_id, recursive)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?

    def unfederate_all_objects(self, ext_dataclay_id):
        session_id = self.get_session_id()
        self.logger.debug("[==UnfederateAllObjects==] Starting unfederation of all objects with dataClay %s, and session %s", ext_dataclay_id, session_id)
        self.ready_clients["@LM"].unfederate_all_objects(session_id, ext_dataclay_id)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?

    def unfederate_all_objects_with_all_dcs(self):
        session_id = self.get_session_id()
        self.logger.debug("[==UnfederateAllObjectsWithAllDCs==] Starting unfederation of all objects with all dcs, and session %s", session_id)
        self.ready_clients["@LM"].unfederate_all_objects_with_all_dcs(session_id)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?
        
    def unfederate_object_with_all_dcs(self, object_id, recursive):
        session_id = self.get_session_id()
        self.logger.debug("[==UnfederateObjectWithAllDCs==] Starting unfederation of object with all dcs, and session %s", session_id)
        self.ready_clients["@LM"].unfederate_object_with_all_dcs(session_id, object_id, recursive)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?
    
    def migrate_federated_objects(self, origin_dataclay_id, dest_dataclay_id):
        session_id = self.get_session_id()
        self.logger.debug("[==MigrateFederatedObjects==] Starting migration of federated objects using session %s", session_id)
        self.ready_clients["@LM"].migrate_federated_objects(session_id, origin_dataclay_id, dest_dataclay_id)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?
    
    def federate_all_objects(self, dest_dataclay_id):
        session_id = self.get_session_id()
        self.logger.debug("[==FederateAllObjects==] Starting federation of all my objects using session %s", session_id)
        self.ready_clients["@LM"].federate_all_objects(session_id, dest_dataclay_id)
        # FIXME: ALIAS CACHE SHOULD BE UPDATED FOR OBJECTS WITH ALIAS REMOVED?
               
    def get_by_alias(self, alias, class_id, safe=True):
        if safe:
            oid, class_id, hint = self.ready_clients["@LM"].get_object_from_alias(self.get_session_id(), alias)
            return self.get_object_by_id(oid, class_id, hint)

        if alias in self.alias_cache :
            oid, class_id, hint = self.alias_cache[alias]
        else:
            oid = self.get_object_id_by_alias(alias)
            class_id = class_id
            hint = self.get_object_location_by_id(oid)
            self.logger.debug("Added alias %s to cache", alias)
            self.alias_cache[alias] = oid, class_id, hint

        return self.get_object_by_id(oid, class_id, hint)
    
    def get_object_id_by_alias(self, alias):
        return uuid.uuid5(uuid.NAMESPACE_OID, alias)
    
    def get_object_location_by_id(self, object_id):
        exec_envs = list(self.get_execution_environments_info())
        return exec_envs[hash(object_id) % len(exec_envs)]
    
    def get_object_location_by_alias(self, alias):
        return self.get_object_location_by_id(self.get_object_id_by_alias(alias))
    
    def delete_alias(self, alias):
        self.ready_clients["@LM"].delete_alias(self.get_session_id(), alias)
        self.logger.debug("Removing from cache alias %s", alias)     
        if alias in self.alias_cache :   
            del self.alias_cache[alias]
    
    def new_replica(self, object_id, class_id, hint, backend_id, recursive):
        self.logger.debug("Starting new replica")
        session_id = self.get_session_id()
        DataClayRuntime.ensure_object_is_registered(self, session_id, object_id, class_id, hint)
        result = self.ready_clients["@LM"].new_replica(session_id, object_id, backend_id, recursive)
        if object_id in self.metadata_cache :   
            del self.metadata_cache[object_id]
        return result
    
    def new_version(self, object_id, class_id, hint, backend_id):
        self.logger.debug("Starting new version")
        session_id = self.get_session_id()
        DataClayRuntime.ensure_object_is_registered(self, session_id, object_id, class_id, hint)
        result = self.ready_clients["@LM"].new_version(session_id, object_id, backend_id)
        if object_id in self.metadata_cache :   
            del self.metadata_cache[object_id]
        return result
    
    def consolidate_version(self, version_info):
        session_id = self.get_session_id()
        return self.ready_clients["@LM"].consolidate_version(session_id, version_info)
    
    def move_object(self, instance, source_backend_id, dest_backend_id, recursive):
        self.logger.debug("Moving object %r from %s to %s",
                     instance, source_backend_id, dest_backend_id)
        object_id = instance.get_object_id()
        moved_objs = self.ready_clients["@LM"].move_object(self.get_session_id(), object_id,
                            source_backend_id, dest_backend_id, recursive)
        for oid in moved_objs:
            if oid in self.metadata_cache :   
                del self.metadata_cache[oid]
    
    def exists(self, object_id):
        return self.dataclay_heap_manager.exists_in_heap(object_id)
    
    def register_external_dataclay(self, exthostname, extport):
        """ Register external dataClay for federation
        :param exthostname: external dataClay host name
        :param extport: external dataClay port
        :return: external dataClay ID registered
        :type exthostname: string
        :type extport: int
        :rtype: UUID
        """
        return self.ready_clients["@LM"].register_external_dataclay(exthostname, extport)
    
    def get_or_new_persistent_instance(self, object_id, metaclass_id, hint):
        """ Check if object with ID provided exists in dataClay heap.
        If so, return it. Otherwise, create it.
        :param object_id: ID of object to get or create
        :param metaclass_id: ID of class of the object (needed for creating it)
        :param hint: Hint of the object, can be None.
        :returns: return the instance with object id provided
        :type object_id: ObjectID
        :type metaclass_id: MetaClassID
        :type hint: BackendID
        :rtype: DataClayObject
        """
        if metaclass_id is None:
            metadata = self.ready_clients["@LM"].get_metadata_by_oid(
                self.get_session_id(), object_id)
            metaclass_id = metadata.metaclassID

        return self.dataclay_object_loader.get_or_new_persistent_instance(metaclass_id, object_id, hint)
    
    def get_dataclay_id(self):
        """ Get dataClay ID of current dataClay
        :return: ID of current dataclay
        :rtype: UUID
        """
        return self.ready_clients["@LM"].get_dataclay_id()
    
    def get_external_dataclay_id(self, exthostname, extport):
        """ Get external dataClay ID with host and port identified
        :param exthostname: external dataClay host name
        :param extport: external dataClay port
        :return: None
        :type exthostname: string
        :type extport: int
        :rtype: None
        """
        return self.ready_clients["@LM"].get_external_dataclay_id(exthostname, extport)
    
    def get_external_dataclay_info(self, dataclay_id):
        """ Get external dataClay information
        :param dataclay_id: external dataClay ID
        :return: DataClayInstance information
        :type dataclay_id: UUID
        :rtype: DataClayInstance
        """
        return self.ready_clients["@LM"].get_external_dataclay_info(dataclay_id)
    
    def get_dataclays_object_is_federated_with(self, object_id):
        """ Retrieve dataClay instance ids where the object is federated
        :param object_id: ID of the object
        :return: dataClay instance ids where this object is federated
        :type object_id: UUID
        :rtype: set of UUID
        """
        return self.ready_clients["@LM"].get_dataclays_object_is_federated_with(object_id)
    
    def get_external_source_of_dataclay_object(self, object_id):
        """ Retrieve dataClay instance id where the object comes from or NULL
        :param object_id: ID of the object
        :return: dataClay instance ids where this object is federated
        :type object_id: UUID
        :rtype: UUID
        """
        return self.ready_clients["@LM"].get_external_source_of_dataclay_object(object_id)

    def get_location(self, object_id):
        self.logger.debug("Looking for location of object %s", str(object_id))
        try:
            obj = self.get_from_heap(object_id)
            if obj is not None:
                hint = obj.get_hint()
                if hint is not None:
                    self.logger.debug("Returning hint from heap object")
                    return hint
                else:
                    raise DataClayException("The object %s is not initialized well, hint missing or not exist" % object_id)
            else:
                raise DataClayException("The object %s is not initialized" % object_id)
        except DataClayException as e:
            # If the object is not initialized well trying to obtain location from metadata
            metadata = self.get_metadata(object_id)
            return six.advance_iterator(iter(metadata.locations))

    def get_metadata(self, object_id):
        if object_id in self.metadata_cache:
            self.logger.debug("Object %s metadata found in cache", object_id)
            metadata = self.metadata_cache[object_id]
        else:
            metadata = self.ready_clients["@LM"].get_metadata_by_oid(
                    self.get_session_id(), object_id)
            if metadata is None: 
                self.logger.debug("Object %s not registered", object_id)
                raise DataClayException("The object %s is not registered" % object_id)
            self.metadata_cache[object_id] = metadata
            self.logger.debug("Received the following MetaDataInfo for object %s: %s",
                    object_id, metadata)
        return metadata

    def get_all_locations(self, object_id):
        self.logger.debug("Getting all locations of object %s", object_id)
        try:
            metadata = self.get_metadata(object_id)
            return metadata.locations
        except:
            self.logger.debug("Object %s has not metadata", object_id)
            obj = self.get_from_heap(object_id)
            if obj is not None:
                hint = obj.get_hint()
                if hint is not None:
                    self.logger.debug("Returning list with hint from heap object")
                    locations = dict()
                    locations[hint] = self.get_execution_environments_info()[hint]
                    return locations
                else:
                    raise DataClayException("The object %s is not initialized well, hint missing or not exist" % object_id)
            else:
                raise DataClayException("The object %s is not initialized" % object_id)

    def get_execution_environments_info(self):
        if self.ee_info_map is None:
            self.ee_info_map = self.ready_clients["@LM"].get_execution_environments_info(self.get_session_id(), LANG_PYTHON,
                                                                                         not self.is_exec_env())
            if self.logger.isEnabledFor(TRACE):
                n = len(self.ee_info_map)
                self.logger.trace("Response of ExecutionEnvironmentsInfo returned #%d ExecutionEnvironmentsInfo", n)
                for i, (ee_id, ee_info) in enumerate(self.ee_info_map.items(), 1):
                    self.logger.trace("ExecutionEnvironments info (#%d/%d): %s\n%s", i, n, ee_id, ee_info)
                self.logger.trace("Response of ExecutionEnvironmentsInfo returned #%d ExecutionEnvironmentsInfo", len(self.ee_info_map))

        return self.ee_info_map

    def get_execution_environments_names(self, username, password):
        credential = (None, password)
        user_id = self.ready_clients["@LM"].get_account_id(username)
        ee_names = self.ready_clients["@LM"].get_execution_environments_names(user_id, credential, LANG_PYTHON)
        if self.logger.isEnabledFor(TRACE):
            n = len(ee_names)
            self.logger.trace("Response of ExecutionEnvironmentsInfo returned #%d ExecutionEnvironmentsInfo", n)
            
        return ee_names

    def activate_tracing(self):
        """ Activate tracing """ 
        if get_current_available_task_id() == 0:
            # Task id 0 means that we are not using compss and extrae must be initialized on client
            initialize_extrae(initialize=True) 
        else:
            initialize_extrae()

    def deactivate_tracing(self):
        """Close the runtime paraver manager and deactivate the traces in LM (That deactivate also the DS)"""
        if get_task_id() == 0:
            finish_tracing(finalize_extrae=True)
        else:
            finish_tracing()

    def activate_tracing_in_dataclay_services(self):
        """Activate the traces in LM (That activate also the DS) """
        if extrae_tracing_is_enabled():
            self.ready_clients["@LM"].activate_tracing(get_current_available_task_id())

    def deactivate_tracing_in_dataclay_services(self):
        """Deactivate the traces in LM and DSs """
        if extrae_tracing_is_enabled():
            self.ready_clients["@LM"].deactivate_tracing()

    def get_traces_in_dataclay_services(self):
        """ Get temporary traces from LM and DSs and store it in current workspace """
        traces = self.ready_clients["@LM"].get_traces()
        traces_dir = Configuration.TRACES_DEST_PATH
        if len(traces) > 0:
            set_path = Configuration.TRACES_DEST_PATH + "/set-0"
            trace_mpits = Configuration.TRACES_DEST_PATH + "/TRACE.mpits"
            with open(trace_mpits, "a+") as trace_file:
                # store them here 
                for key, value in traces.items():
                    dest_path = set_path + "/" + key
                    self.logger.debug("Storing object %s" % dest_path)
                    with open(dest_path, "wb") as dest_file:
                        dest_file.write(value)
                        dest_file.close() 
                    if key.endswith("mpit"):
                        pointer = str(dest_path) + " named\n"
                        trace_file.write(pointer)
                        self.logger.debug("Appending to %s line: %s" % (trace_mpits, pointer))
                trace_file.flush()
                trace_file.close()

    def stop_gc(self):
        """
        @postcondition: stop GC. useful for shutdown. 
        """ 
        # Stop HeapManager
        self.logger.debug("Stopping GC. Sending shutdown event.")
        self.dataclay_heap_manager.shutdown()
        self.logger.debug("Waiting for GC.")
        self.dataclay_heap_manager.join()
        self.logger.debug("GC stopped.")

    def stop_runtime(self):
        """ 
        @postcondition: Stop connections and daemon threads. 
        """ 
        self.logger.verbose("** Stopping runtime **")

        for name, client in self.ready_clients.items():
            self.logger.verbose("Closing client connection to %s", name)
            client.close()
        
        self.ready_clients = {}
        
        # Stop HeapManager
        self.stop_gc()
        
    ################## EXTRAE IGNORED FUNCTIONS ###########################
    deactivate_tracing.do_not_trace = True
    activate_tracing.do_not_trace = True
    activate_tracing_in_dataclay_services.do_not_trace = True 
    deactivate_tracing_in_dataclay_services.do_not_trace = True 
        



"""Custom container for holding extra information on classes.

The dataClay friendly classed (either before registration or stub-generated
ones) will contain their extra information --like the full_name, the class_id
or the namespace-- in DataClayExtraData instances.
"""

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class DataClayCommonExtraData(object):
    """ExtraData commonlib base class.

    ExtraData behaves as a dictionary, with some optional extras.
    """
    __slots__ = list()

    def __init__(self, **kwargs):
        # load values into slots
        for k, v in kwargs.items():
            # Let the errors trigger from here (if kwargs is invalid)
            setattr(self, k, v)

        # force None (to avoid attribute errors)
        for s in self.__slots__:
            try:
                _ = getattr(self, s)
            except AttributeError:
                setattr(self, s, None)

    def __str__(self):
        ret = ["{%s contents:" % self.__class__.__name__, ]

        for s in self.__slots__:
            ret.append("  %s: %s" % (s, getattr(self, s)))
        ret.append("}")

        return "\n".join(ret)


class DataClayClassExtraData(DataClayCommonExtraData):
    """Container for ExtraData related to a certain dataClay Class

    Instances for this class are typically associated to DataClayObject
    derived classes (and automatically populated by the ExecutionGateway).
    """
    __slots__ = ('full_name',
                 'namespace',
                 'classname',
                 'class_id',
                 'properties',
                 'imports',
                 'stub_info',
                 'metaclass_container',
                 )


class DataClayInstanceExtraData(DataClayCommonExtraData):
    """Container for ExtraData related to a certain dataClay Object instance.

    Instances for this class are typically created and assigned to a certain
    DataClayObject instance when creating them. The ExecutionGateway populates
    its data.
    """
    __slots__ = ('persistent_flag',
                 'object_id',
                 'master_location',
                 'dataset_id',
                 'execenv_id',
                 'loaded_flag',
                 'pending_to_register_flag',
                 'owner_session_id',
                 'dirty_flag',
                 'memory_pinned',
                 )




""" Class description goes here. """

from decorator import decorate
import logging
import six
import traceback
from dataclay.commonruntime.Runtime import getRuntime

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


def _dclayMethod(f, self, *args, **kwargs):
    """Helper function for DataClayObject method decoration"""
    logger.verbose("Calling function %s", f.__name__)
    is_exec_env = getRuntime().is_exec_env()
    try:
        if (is_exec_env and self.is_loaded()) \
                or (not is_exec_env and not self.is_persistent())\
                or f._dclay_local:
            return f(self, *args, **kwargs)
        else:
            return getRuntime().execute_implementation_aux(f.__name__, self, args, self.get_hint())
    except Exception:
        traceback.print_exc()
        raise


def _dclayEmptyMethod(f, self, *args, **kwargs):
    """Similar to dclayMethod, but without actual Python implementation."""
    logger.verbose("Calling (languageless) function %s", f.__name__)

    # Let it fail elsewhere, if the user hacks around into an invalid state
    # (like a loaded&local non-persistent instance with an dclayEmptyMethod,
    #  something that should not happen normally)
    return getRuntime().execute_implementation_aux(f.__name__, self, args, self.get_hint())


class dclayMethod(object):
    """Class-based decorator for DataClayObject method decoration"""

    def __init__(self, **kwargs):
        """Provide the argument type information

        The programmer is expected to set the same kwargs as the function,
        in addition to the `return_` special method return type.

        The typical usage is:

            @dclayMethod(a=int,
                         b='some.path.to.Class',  # this is valid (str)
                         c=imported.path.to.Class, # this is also valid
                         return_=None)
            def foo_bar(a, b, c):
                ...

        The method should be inside a DataClayObject derived class. See both
        DataClayObject class implementation and ExecutionGateway metaclass for
        more information about the internal behaviour.
        """
        self._method_args = kwargs

    def __call__(self, f):
        if six.PY2:
            logger.verbose('Preparing dataClay method `%s` with arguments: %s',
                     f.func_name, self._method_args)
        elif six.PY3:
            logger.verbose('Preparing dataClay method `%s` with arguments: %s',
                     f.__name__, self._method_args)
        decorated_func = decorate(f, _dclayMethod)
        decorated_func._dclay_entrypoint = f
        decorated_func._dclay_ret = self._method_args.pop("return_", None)
        decorated_func._dclay_args = self._method_args
        decorated_func._dclay_method = True

        # Store the local flag in both the function and the decorated function
        is_local = self._method_args.pop("_local", False)
        f._dclay_local = is_local
        # TODO: at the moment, this flag is ignored
        decorated_func._dclay_local = is_local

        # ToDo: something more clever, or add the user-provided flag option at least
        decorated_func._dclay_readonly = False
        return decorated_func


def dclayEmptyMethod(f):
    """Simple (parameter-less) decorator for languageless methods."""
    decorated_func = decorate(f, _dclayEmptyMethod)
    decorated_func._dclay_method = True
    # ToDo: something more clever, or add the user-provided flag option at least
    decorated_func._dclay_readonly = False
    return decorated_func




""" Class description goes here. """

from dataclay.exceptions.ErrorDefs import ErrorCodes

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class DataClayException(Exception):
    """Base class for exceptions in this module."""
    pass


class ImproperlyConfigured(DataClayException):
    """Raised when the settings are not well-formed."""
    # def __init__(self, msg):
    #     self.msg = msg
    pass


class IdentifierNotFound(DataClayException):
    """Raised when a certain identifier (UUID, name...) has not been found."""
    pass


class InvalidPythonSignature(DataClayException):
    """Raised when trying to use a not recognizable Python-signature."""
    pass


class RemoteException(RuntimeError):
    """Exception thrown in client code after a RPC call return with an exception."""

    def __init__(self, error_code, error_string):
        self.error_code = error_code
        self.error_string = error_string
        try:
            self.error_name = ErrorCodes.error_codes[error_code]
        except KeyError:
            self.error_name = "UNDEFINED".format(error_code)
        super(RuntimeError, self).__init__("Error [{}: {}]. Server response: {}".format(
            self.error_code, self.error_name, self.error_string))


class NetworkError(RuntimeError):
    """Exception when some socket input/output recv or similar operation
    does not behave as expected."""

    def __init__(self, *args):
        super(RuntimeError, self).__init__(*args)


class ClientError(RuntimeError):
    """Exception when a client has sent some invalid request."""

    def __init__(self, *args):
        super(RuntimeError, self).__init__(*args)




""" Class description goes here. """

"""Package containing server gRPC classes."""

__author__ = 'Enrico La Sala <enrico.lasala@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'




""" Class description goes here. """

"""Generic GRPC paraver interceptor mechanisms.

In general, the source of examples for making this work comes from:

Heavily inspired in:
https://github.com/grpc/grpc/blob/master/examples/python/interceptors/
"""

HEADER_MESSAGEID = "messageid"
HEADER_CLIENTPORT = "clientport"




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class SessionImplementation(ManagementObject):
    _fields = ["id",
               "implementationID",
               "namespaceID",
               "respAccountID",
               ]

    _internal_fields = []




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject

    
class SessionProperty(ManagementObject):
    _fields = ["id",
               "propertyID",
               ]

    _internal_fields = []

